""" Base GNodeSettings """

from gwbase.config.g_node_settings import GNodeSettings
from gwbase.config.g_node_settings import SupervisorSettings


__all__ = [
    "GNodeSettings",
    "SupervisorSettings",
]



"""Type heartbeat.a, version 001 - sample of type evolution with versions."""

import json
from typing import Any
from typing import Dict
from typing import Literal

from gw.errors import GwTypeError
from pydantic import BaseModel
from pydantic import field_validator

from pydantic.alias_generators import to_pascal
from pydantic.alias_generators import to_snake

from gwbase.types import HeartbeatA

class HeartbeatA001(BaseModel):
    """
    This earlier version of type heartbeat.a does not enforce that MyHex and
    YourLastHex are single character hexes - just that they are strings
    """
    my_hex: str  #
    your_last_hex: str  #
    type_name: Literal["heartbeat.a"] = "heartbeat.a"
    version: str = "001"

    class Config:
        populate_by_name = True
        alias_generator = to_pascal

    def as_dict(self) -> Dict[str, Any]:
        d = {
            to_pascal(key): value
            for key, value in self.model_dump().items()
            if value is not None
        }
        return d


    def as_type(self) -> str:
        return json.dumps(self.as_dict())


class HeartbeatA001_Maker:
    type_name = "heartbeat.a"
    version = "001"


    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA001) -> str:
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> HeartbeatA001:
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA001:
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing MyHex")
        if "YourLastHex" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing YourLastHex")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"dict {d2} missing TypeName")
        d3 = {to_snake(key): value for key, value in d2.items()}
        return HeartbeatA001(**d3)




"""Type heartbeat.a, version 100.
Change since 001: enforces that MyHex and YourLastHex are hex characters
"""

import json
import logging
from typing import Any
from typing import Dict
from typing import Literal

from gw.errors import GwTypeError
from gw.utils import is_pascal_case
from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator
from pydantic.alias_generators import to_pascal
from pydantic.alias_generators import to_snake


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class HeartbeatA(BaseModel):
    """
    Used to check that an actor can both send and receive messages.

    Payload for direct messages sent back and forth between actors, for example a Supervisor
    and one of its subordinates. Update: MyHex and YourLatestHex must have format HexChar

    [More info](https://gridworks.readthedocs.io/en/latest/g-node-instance.html)
    """

    my_hex: str = Field(
        title="Hex character getting sent",
        default="0",
    )
    your_last_hex: str = Field(
        title="Last hex character received from heartbeat partner",
        default="0",
    )
    type_name: Literal["heartbeat.a"] = "heartbeat.a"
    version: Literal["100"] = "100"

    class Config:
        populate_by_name = True
        alias_generator = to_pascal

    @field_validator("my_hex")
    def _check_my_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"MyHex failed HexChar format validation: {e}")
        return v

    @field_validator("your_last_hex")
    def _check_your_last_hex(cls, v: str) -> str:
        try:
            check_is_hex_char(v)
        except ValueError as e:
            raise ValueError(f"YourLastHex failed HexChar format validation: {e}")
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        heartbeat.a.100 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        heartbeat.a.100 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            to_pascal(key): value
            for key, value in self.model_dump().items()
            if value is not None
        }
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the heartbeat.a.100 representation.

        Instances in the class are python-native representations of heartbeat.a.100
        objects, while the actual heartbeat.a.100 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is HeartbeatA.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class HeartbeatA_Maker:
    type_name = "heartbeat.a"
    version = "100"

    @classmethod
    def tuple_to_type(cls, tuple: HeartbeatA) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> HeartbeatA:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> HeartbeatA:
        """
        Deserialize a dictionary representation of a heartbeat.a.100 message object
        into a HeartbeatA python object for internal use.

        This is the near-inverse of the HeartbeatA.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a HeartbeatA object.

        Returns:
            HeartbeatA
        """
        for key in d.keys():
            if not is_pascal_case(key):
                raise GwTypeError(f"Key '{key}' is not PascalCase")
        d2 = dict(d)
        if "MyHex" not in d2.keys():
            raise GwTypeError(f"dict missing MyHex: <{d2}>")
        if "YourLastHex" not in d2.keys():
            raise GwTypeError(f"dict missing YourLastHex: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "100":
            LOGGER.debug(
                f"Attempting to interpret heartbeat.a version {d2['Version']} as version 100"
            )
            d2["Version"] = "100"
        d3 = {to_snake(key): value for key, value in d2.items()}
        return HeartbeatA(**d3)


def check_is_hex_char(v: str) -> None:
    """Checks HexChar format

    HexChar format: single-char string in '0123456789abcdefABCDEF'

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not HexChar format
    """
    if not isinstance(v, str):
        raise ValueError(f"<{v}> must be a hex char, but not even a string")
    if len(v) > 1:
        raise ValueError(f"<{v}> must be a hex char, but not of len 1")
    if v not in "0123456789abcdefABCDEF":
        raise ValueError(f"<{v}> must be one of '0123456789abcdefABCDEF'")



import algosdk

from gwbase.enums import UniverseType


def check_is_algo_secret_key_format(v: str) -> None:
    try:
        b = algosdk.account.address_from_private_key(v)
    except Exception as e:
        raise ValueError(
            f"Not Algorand Secret Key format! Generate one by: \n"
            f"import algosdk\n"
            "private_key, address = algosdk.account.generate_account()\n"
            f"incorrect sk: {v}"
        )


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_reasonable_unix_time_s(v: int) -> None:
    """
    ReasonableUnixTimeS format: time in unix seconds between Jan 1 2000 and Jan 1 3000

    Raises:
        ValueError: if not ReasonableUnixTimeS format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp = int(start_date.timestamp())
    end_timestamp = int(end_date.timestamp())

    if v < start_timestamp:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp:
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_g_node_alias(alias: str, universe_type: UniverseType) -> None:
    check_is_left_right_dot(alias)
    first_word = alias.split(".")[0]
    if universe_type == UniverseType.Dev:
        if not first_word.startswith("d"):
            raise ValueError(
                f"Bad alias {alias} ... Dev universe GNode alias must start with d."
            )
    elif universe_type == UniverseType.Hybrid:
        if not first_word.startswith("h"):
            raise ValueError(
                f"Bad alias {alias} ... Hybrid universe GNode alias must start with h."
            )
    elif universe_type == UniverseType.Production:
        if not first_word == "w":
            raise ValueError(
                f"Bad alias {alias} ... Production universe GNode alias must have w as first word."
            )



from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class UniverseType(GwStrEnum):
    """
    Allows for multiple GridWorks, in particular for development and shared simulations

    Enum universe.type version 000 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#universetype)
      - [More Info](https://gridworks.readthedocs.io/en/latest/universe.html)

    Values (with symbols in parens):
      - Dev (00000000): Simulation running on a single computer.
      - Hybrid (8281a0fa): Anything goes.
      - Production (b0168bda): Money at stake.
    """

    Dev = auto()
    Hybrid = auto()
    Production = auto()

    @classmethod
    def default(cls) -> "UniverseType":
        """
        Returns default value (in this case Dev)
        """
        return cls.Dev

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "000"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (universe.type)
        """
        return "universe.type"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (000)
        """
        return "000"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "Dev".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a UniverseType enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "8281a0fa",
            "b0168bda",
        ]


symbol_to_value = {
    "00000000": "Dev",
    "8281a0fa": "Hybrid",
    "b0168bda": "Production",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "Dev": "000",
    "Hybrid": "000",
    "Production": "000",
}



from gwbase_test.stub_actors import GNodeStubRecorder
from gwbase_test.stub_actors import SupervisorStubRecorder
from gwbase_test.stub_actors import TimeCoordinatorStubRecorder
from gwbase_test.stub_actors import load_rabbit_exchange_bindings


__all__ = [
    "load_rabbit_exchange_bindings",
    "GNodeStubRecorder",
    "SupervisorStubRecorder",
    "TimeCoordinatorStubRecorder",
]



from enum import auto
from typing import List
from typing import Optional

from gw.enums import GwStrEnum


class GNodeRole(GwStrEnum):
    """
    Categorizes GNodes by their function within GridWorks

    Enum g.node.role version 001 in the GridWorks Type registry.

    Used by used by multiple Application Shared Languages (ASLs). For more information:
      - [ASLs](https://gridworks-type-registry.readthedocs.io/en/latest/)
      - [Global Authority](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#gnoderole)
      - [More Info](https://gridworks.readthedocs.io/en/latest/g-node-role.html)

    Values (with symbols in parens):
      - GNode (00000000): Default value
      - TerminalAsset (bdeaa0b1): An avatar for a real-word Transactive Device [More Info](https://gridworks.readthedocs.io/en/latest/transactive-device.html).
      - AtomicTNode (8021dcad): Transacts in markets on behalf of, and controlling the power use
        of, a TerminalAsset [More Info](https://gridworks.readthedocs.io/en/latest/atomic-t-node.html).
      - MarketMaker (304890c5): Runs energy markets at its Node in the GNodeTree [More Info](https://gridworks.readthedocs.io/en/latest/market-maker.html).
      - AtomicMeteringNode (8eb5b9e1): Role of a GNode that will become an AtomicTNode, prior to
        it owning TaTradingRights
      - ConductorTopologyNode (234cfaa2): An avatar for a real-world electric grid node - e.g. a
        substation or transformer
      - InterconnectionComponent (fec0c127): An avatar for a cable or wire on the electric grid
      - World (3901c7d2): Adminstrative GNode responsible for managing and authorizing instances [More Info](https://gridworks.readthedocs.io/en/latest/world-role.html).
      - TimeCoordinator (c499943c): Responsible for managing time in simulations
      - Supervisor (88112a93): Responsible for GNode actors running in a container [More Info](https://gridworks.readthedocs.io/en/latest/supervisor.html).
      - Scada (674ad859): GNode associated to the device and code that directly monitors and actuates
        a Transactive Device
      - PriceService (2161739f): Provides price forecasts for markets run by MarketMakers
      - WeatherService (1dce1efd): Provides weather forecasts
      - AggregatedTNode (db57d184): An aggregation of AtomicTNodes
      - Persister (07f28817): Responsible for acking events with delivery guarantees
    """

    GNode = auto()
    TerminalAsset = auto()
    AtomicTNode = auto()
    MarketMaker = auto()
    AtomicMeteringNode = auto()
    ConductorTopologyNode = auto()
    InterconnectionComponent = auto()
    World = auto()
    TimeCoordinator = auto()
    Supervisor = auto()
    Scada = auto()
    PriceService = auto()
    WeatherService = auto()
    AggregatedTNode = auto()
    Persister = auto()

    @classmethod
    def default(cls) -> "GNodeRole":
        """
        Returns default value (in this case GNode)
        """
        return cls.GNode

    @classmethod
    def values(cls) -> List[str]:
        """
        Returns enum choices
        """
        return [elt.value for elt in cls]

    @classmethod
    def version(cls, value: Optional[str] = None) -> str:
        """
        Returns the version of the class (default) used by this package or the
        version of a candidate enum value (always less than or equal to the version
        of the class)

        Args:
            value (Optional[str]): None (for version of the Enum itself) or
            the candidate enum value.

        Raises:
            ValueError: If the value is not one of the enum values.

        Returns:
            str: The version of the enum used by this code (if given no
            value) OR the earliest version of the enum containing the value.
        """
        if value is None:
            return "001"
        if not isinstance(value, str):
            raise ValueError(f"This method applies to strings, not enums")
        if value not in value_to_version.keys():
            raise ValueError(f"Unknown enum value: {value}")
        return value_to_version[value]

    @classmethod
    def enum_name(cls) -> str:
        """
        The name in the GridWorks Type Registry (g.node.role)
        """
        return "g.node.role"

    @classmethod
    def enum_version(cls) -> str:
        """
        The version in the GridWorks Type Registry (001)
        """
        return "001"

    @classmethod
    def symbol_to_value(cls, symbol: str) -> str:
        """
        Given the symbol sent in a serialized message, returns the encoded enum.

        Args:
            symbol (str): The candidate symbol.

        Returns:
            str: The encoded value associated to that symbol. If the symbol is not
            recognized - which could happen if the actor making the symbol is using
            a later version of this enum, returns the default value of "GNode".
        """
        if symbol not in symbol_to_value.keys():
            return cls.default().value
        return symbol_to_value[symbol]

    @classmethod
    def value_to_symbol(cls, value: str) -> str:
        """
        Provides the encoding symbol for a GNodeRole enum to send in seriliazed messages.

        Args:
            symbol (str): The candidate value.

        Returns:
            str: The symbol encoding that value. If the value is not recognized -
            which could happen if the actor making the message used a later version
            of this enum than the actor decoding the message, returns the default
            symbol of "00000000".
        """
        if value not in value_to_symbol.keys():
            return value_to_symbol[cls.default().value]
        return value_to_symbol[value]

    @classmethod
    def symbols(cls) -> List[str]:
        """
        Returns a list of the enum symbols
        """
        return [
            "00000000",
            "bdeaa0b1",
            "8021dcad",
            "304890c5",
            "8eb5b9e1",
            "234cfaa2",
            "fec0c127",
            "3901c7d2",
            "c499943c",
            "88112a93",
            "674ad859",
            "2161739f",
            "1dce1efd",
            "db57d184",
            "07f28817",
        ]


symbol_to_value = {
    "00000000": "GNode",
    "bdeaa0b1": "TerminalAsset",
    "8021dcad": "AtomicTNode",
    "304890c5": "MarketMaker",
    "8eb5b9e1": "AtomicMeteringNode",
    "234cfaa2": "ConductorTopologyNode",
    "fec0c127": "InterconnectionComponent",
    "3901c7d2": "World",
    "c499943c": "TimeCoordinator",
    "88112a93": "Supervisor",
    "674ad859": "Scada",
    "2161739f": "PriceService",
    "1dce1efd": "WeatherService",
    "db57d184": "AggregatedTNode",
    "07f28817": "Persister",
}

value_to_symbol = {value: key for key, value in symbol_to_value.items()}

value_to_version = {
    "GNode": "000",
    "TerminalAsset": "000",
    "AtomicTNode": "000",
    "MarketMaker": "000",
    "AtomicMeteringNode": "000",
    "ConductorTopologyNode": "000",
    "InterconnectionComponent": "000",
    "World": "000",
    "TimeCoordinator": "000",
    "Supervisor": "000",
    "Scada": "000",
    "PriceService": "000",
    "WeatherService": "000",
    "AggregatedTNode": "000",
    "Persister": "001",
}



""" List of all the types """

from gwbase.types.base_g_node_gt import BaseGNodeGt
from gwbase.types.base_g_node_gt import BaseGNodeGt_Maker
from gwbase.types.g_node_gt import GNodeGt
from gwbase.types.g_node_gt import GNodeGt_Maker
from gwbase.types.g_node_instance_gt import GNodeInstanceGt
from gwbase.types.g_node_instance_gt import GNodeInstanceGt_Maker
from gwbase.types.heartbeat_a import HeartbeatA
from gwbase.types.heartbeat_a import HeartbeatA_Maker
from gwbase.types.ready import Ready
from gwbase.types.ready import Ready_Maker
from gwbase.types.sim_timestep import SimTimestep
from gwbase.types.sim_timestep import SimTimestep_Maker
from gwbase.types.super_starter import SuperStarter
from gwbase.types.super_starter import SuperStarter_Maker
from gwbase.types.supervisor_container_gt import SupervisorContainerGt
from gwbase.types.supervisor_container_gt import SupervisorContainerGt_Maker


__all__ = [
    "BaseGNodeGt",
    "BaseGNodeGt_Maker",
    "GNodeGt",
    "GNodeGt_Maker",
    "GNodeInstanceGt",
    "GNodeInstanceGt_Maker",
    "HeartbeatA",
    "HeartbeatA_Maker",
    "Ready",
    "Ready_Maker",
    "SimTimestep",
    "SimTimestep_Maker",
    "SuperStarter",
    "SuperStarter_Maker",
    "SupervisorContainerGt",
    "SupervisorContainerGt_Maker",
]



from pydantic import BaseModel
from pydantic import SecretStr


class RabbitBrokerClient(BaseModel):
    """Settings for connecting to a Rabbit Broker"""

    url: SecretStr = SecretStr("amqp://smqPublic:smqPublic@localhost:5672/d1__1")



import copy
import datetime
import time
from typing import List
from typing import Optional

from pika.channel import Channel as PikaChannel
from pydantic import BaseModel

from gwbase.actor_base import ActorBase  # TODO: from gwatn import ActorBase
from gwbase.actor_base import RabbitRole
from gwbase.config import GNodeSettings
from gwbase.enums import GNodeRole
from gwbase.types import HeartbeatA
from gwbase.types import Ready
from gwbase.types import Ready_Maker


class ExchangeBinding(BaseModel):
    From: str
    To: str
    Key: str


def load_rabbit_exchange_bindings(ch: PikaChannel) -> None:
    """
    Creation of exchanges and bindings for use in rabbit broker;
    useful if/when the definitions cannot be imported via json in ci
    - which happened for us with github actions
    :param PikaChannel ch: any open channel to the rabbit broker
    """
    if ch is None:
        raise Exception(f"Channel is None! Make sure you have started the actor")
    if not ch.is_open:
        raise Exception(
            f"Channel is not open yet! Make sure the channel is open before calling"
        )

    ch.exchange_declare(
        exchange="ear_tx",
        exchange_type="topic",
        durable=True,
        internal=True,
    )

    for role in RabbitRole.values():
        ch.exchange_declare(
            exchange=f"{role}_tx",
            exchange_type="topic",
            durable=True,
            internal=True,
        )
        ch.exchange_declare(
            exchange=f"{role}mic_tx",
            exchange_type="topic",
            durable=True,
            internal=False,
        )

    bindings: List[ExchangeBinding] = [
        ExchangeBinding(From="atomictnode", To="ear", Key="#"),
        ExchangeBinding(
            From="atomictnode",
            To="supervisor",
            Key="*.*.atomictnode.*.supervisor.*",
        ),
        ExchangeBinding(
            From="atomictnode",
            To="timecoordinator",
            Key="*.*.atomictnode.*.timecoordinator.*",
        ),
        ExchangeBinding(From="gnode", To="ear", Key="#"),
        ExchangeBinding(From="gnode", To="supervisor", Key="*.*.gnode.*.supervisor.*"),
        ExchangeBinding(
            From="gnode", To="timecoordinator", Key="*.*.gnode.*.timecoordinator.*"
        ),
        ExchangeBinding(From="marketmaker", To="ear", Key="#"),
        ExchangeBinding(
            From="marketmaker",
            To="supervisor",
            Key="*.*.marketmaker.*.supervisor.*",
        ),
        ExchangeBinding(
            From="marketmaker",
            To="timecoordinator",
            Key="*.*.marketmaker.*.timecoordinator.*",
        ),
        ExchangeBinding(From="supervisor", To="ear", Key="#"),
        ExchangeBinding(
            From="supervisor",
            To="atomictnode",
            Key="*.*.supervisor.*.atomictnode.*",
        ),
        ExchangeBinding(
            From="supervisor",
            To="marketmaker",
            Key="*.*.supervisor.*.marketmaker.*",
        ),
        ExchangeBinding(
            From="supervisor", To="supervisor", Key="*.*.supervisor*.supervisor.*"
        ),
        ExchangeBinding(
            From="supervisor",
            To="timecoordinator",
            Key="*.*.supervisor.*.timecoordinator.*",
        ),
        ExchangeBinding(From="world", To="ear", Key="#"),
        ExchangeBinding(
            From="world", To="timecoordinator", Key="*.*.world.*.timecoordinator.*"
        ),
    ]

    for binding in bindings:
        ch.exchange_bind(
            destination=f"{binding.To}_tx",
            source=f"{binding.From}mic_tx",
            routing_key=binding.Key,
        )


class GNodeStubRecorder(ActorBase):
    messages_received: int = 0
    messages_routed_internally: int = 0
    latest_from_role: Optional[str] = None
    latest_from_alias: Optional[str] = None
    latest_payload: Optional[HeartbeatA] = None
    got_heartbeat_from_super: bool = False

    def __init__(self, settings: GNodeSettings):
        super().__init__(settings=settings)
        self.settings = settings

    def on_message(self, _unused_channel, basic_deliver, properties, body):
        self.messages_received += 1
        super().on_message(_unused_channel, basic_deliver, properties, body)

    def route_message(self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA):
        self.messages_routed_internally += 1
        self.latest_payload = payload
        if isinstance(payload, HeartbeatA):
            self.heartbeat_a_received(from_alias, from_role, payload)

    def heartbeat_a_received(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ):
        if (
            from_alias == self.settings.my_super_alias
            and from_role == GNodeRole.Supervisor
        ):
            self.got_heartbeat_from_super = True

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    def summary_str(self):
        """Summarize results in a string"""
        return (
            f"{self.g_node_role.value} [{self.alias}] messages_received: {self.messages_received}  "
            f"latest_payload: {self.latest_payload}"
        )


class SupervisorStubRecorder(GNodeStubRecorder):
    got_heartbeat_from_sub: bool = False

    def __init__(self, settings: GNodeSettings, subordinate_alias: str):
        my_settings = copy.deepcopy(settings)
        my_settings.g_node_alias = "d1.super"
        my_settings.g_node_role_value = "Supervisor"
        super().__init__(settings=my_settings)
        self.my_single_sub = subordinate_alias

    def heartbeat_a_received(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ):
        """Used to test that a Supervisor gets a message from GNode"""
        if from_alias == self.my_single_sub:
            self.got_heartbeat_from_sub = True


class TimeCoordinatorStubRecorder(GNodeStubRecorder):
    """This stub timecoordinator is intended to be used"""

    def __init__(self, settings: GNodeSettings):
        self._time: int = int(
            datetime.datetime(
                year=2020, month=1, day=1, hour=5, tzinfo=datetime.timezone.utc
            ).timestamp()
        )
        my_settings = copy.deepcopy(settings)
        my_settings.g_node_alias = "d1.time"
        my_settings.g_node_role_value = "TimeCoordinator"
        super().__init__(settings=my_settings)
        self.my_actors: List[str] = [GNodeSettings().g_node_alias]
        self.ready: List[str] = []

    def route_message(self, from_alias: str, from_role: GNodeRole, payload: Ready):
        if payload.TypeName == Ready_Maker.type_name:
            if from_alias in self.my_actors:
                if payload.TimeUnixS == self._time:
                    self.ready.append(from_alias)

    def is_ready(self) -> bool:
        if set(self.ready) == set(self.my_actors):
            return True
        return False

    def prepare_for_death(self):
        self.actor_main_stopped = True



import logging
from typing import Dict
from typing import List
from typing import Optional

from gw.errors import DcError
from gw.errors import GwTypeError

from gwbase.data_classes.gps_point import GpsPoint
from gwbase.enums import CoreGNodeRole
from gwbase.enums import GNodeStatus


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class BaseGNode:
    """
    DataClass corresponding to base.g.node.gt type.

    The global authority for state of BaseGNodes is managed by the GNodeFactory,
    and shared with AtomicTNodes from its World Registry.
    Do not update or create these objects except when starting up and using local
    data from a file (provided by the World to its docker instance) or in a direct
    message (Rest POST or rabbit broadcast) from the World.

    [more info](https://gridworks.readthedocs.io/en/latest/g-node.html#basegnodes)
    """

    by_id: Dict[int, "BaseGNode"] = {}
    by_alias: Dict[str, "BaseGNode"] = {}

    def __new__(cls, g_node_id, *args, **kwargs):
        try:
            return cls.by_id[g_node_id]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[g_node_id] = instance
            return instance

    def __init__(
        self,
        g_node_id: Optional[str] = None,
        alias: Optional[str] = None,
        status: Optional[GNodeStatus] = None,
        role: Optional[CoreGNodeRole] = None,
        g_node_registry_addr: Optional[str] = None,
        prev_alias: Optional[str] = None,
        gps_point_id: Optional[str] = None,
        ownership_deed_id: Optional[int] = None,
        ownership_deed_validator_addr: Optional[str] = None,
        owner_addr: Optional[str] = None,
        daemon_addr: Optional[str] = None,
        trading_rights_id: Optional[int] = None,
        scada_algo_addr: Optional[str] = None,
        scada_cert_id: Optional[int] = None,
    ):
        self.g_node_id = g_node_id
        self.alias = alias
        if not isinstance(status, GNodeStatus):
            raise DcError(f"status {status} must be GNodeStatus, got {type(status)}")
        self.status = status
        if not isinstance(role, CoreGNodeRole):
            raise DcError(f"role {role} must be  CoreGNodeRole, got {type(role)}")
        self.role = role
        self.g_node_registry_addr = g_node_registry_addr
        self.prev_alias = prev_alias
        self.gps_point_id = gps_point_id
        self.ownership_deed_id = ownership_deed_id
        self.ownership_deed_validator_addr = ownership_deed_validator_addr
        self.owner_addr = owner_addr
        self.daemon_addr = daemon_addr
        self.trading_rights_id = trading_rights_id
        self.scada_algo_addr = scada_algo_addr
        self.scada_cert_id = scada_cert_id
        self.__class__.by_alias[self.alias] = self

    def __repr__(self):
        rs = f"GNode Alias: {self.alias}, Role: {self.role.value}, Status: {self.status.value}"
        if self.ownership_deed_id and self.role == CoreGNodeRole.TerminalAsset:
            rs += f", TaDeedIdx: {self.ownership_deed_id}"
        return rs

    def gps_point(self) -> Optional[GpsPoint]:
        if self.gps_point_id is None:
            return None
        return GpsPoint.by_id["gps_point_id"]

    def is_root(self) -> bool:
        alias_list = self.alias.split(".")
        alias_list.pop()
        if len(alias_list) == 0:
            return True
        return False

    def is_copper(self) -> bool:
        """Returns true if role is not other"""
        if self.role == CoreGNodeRole.Other:
            return False
        return True

    @classmethod
    def active_g_nodes(cls) -> List["BaseGNode"]:
        g_nodes = list(BaseGNode.by_alias.values())
        return list(filter(lambda x: x.status == GNodeStatus.Active, g_nodes))

    @classmethod
    def parent_from_alias(cls, alias: str) -> Optional["BaseGNode"]:
        """
        Returns:
            - BaseGNode. If the parent as suggested by the alias exists as an
            Active BaseGNode, returns that.
            - None. If alias is one word long (i.e. root of world)
        """
        alias_list = alias.split(".")
        alias_list.pop()
        parent_alias = ".".join(alias_list)
        if parent_alias in list(map(lambda x: x.alias, cls.active_g_nodes())):
            return BaseGNode.by_alias[parent_alias]
        return None

    def parent(self) -> Optional["BaseGNode"]:
        """
        Raises: DcError if "natural" parent (as suggested by alias) is not Active,
        and either
            - prev_alias is None, OR
            - the parent as suggested by prev_alias is not Active and/or
            does not exist.
        Returns:
            BaseGNode.   Parent BaseGNode
              - If the parent as suggested by the alias exists as an
            Active BaseGNode, returns that ("natural" parent)
              - Else, if the parent as suggested by the prev_alais exists
              as an active BaseGNode, returns that.
            None.
              - If alias is one word long (i.e. root of world)
        """
        if self.is_root():
            return None
        natural_parent = self.parent_from_alias(self.alias)
        if natural_parent is not None:
            return natural_parent

        # alias may point to incorrect parent if getting updated
        if self.prev_alias is None:
            raise DcError(f"error finding parent for {self.alias}!")

        parent_pending_alias_update = self.parent_from_alias(self.prev_alias)
        if parent_pending_alias_update is None:
            raise DcError(f"error finding parent for {self.alias}!")
        return parent_pending_alias_update

    def children(self) -> List["BaseGNode"]:
        """Returns the list of BaseGnodes identifying this node as parent"""
        return list(filter(lambda x: x.parent() == self, BaseGNode.by_alias.values()))

    ##########################
    # Object creation
    #########################

    @classmethod
    def check_creation_axioms(cls, attributes):
        cls._creation_axiom_1(attributes)
        cls._creation_axiom_2(attributes)
        cls._creation_axiom_3(attributes)
        cls._creation_axiom_4(attributes)
        cls._creation_axiom_5(attributes)
        cls._schema_axiom_1(attributes)
        cls._schema_axiom_2(attributes)
        cls._schema_axiom_3(attributes)
        cls._schema_axiom_4(attributes)
        cls._joint_axiom_1(attributes)
        cls._joint_axiom_2(attributes)

    @classmethod
    def _creation_axiom_1(cls, attributes):
        """Creation Axiom 1: alias cannot be used, now or previously, by this or
        other BaseGNodes"""
        # Done in models.py, as it requires access to BaseGNodeHistory

    @classmethod
    def _creation_axiom_2(cls, attributes):
        """Creation Axiom 2: g_node_id must be a unique string of UUID format"""
        g_node_id = attributes["g_node_id"]
        if g_node_id in cls.by_id.keys():
            raise DcError(f"g_node_id {attributes['g_node_id']} already in use")
        if not isinstance(g_node_id, str):
            raise DcError(f"g_node_id must be a string, got {type(g_node_id)}.")
        try:
            check_is_uuid_canonical_textual(g_node_id)
        except GwTypeError:
            raise DcError(
                f"g_node_id must have format UuidCanonicalTextutal. Got {g_node_id}."
            )

    @classmethod
    def _creation_axiom_3(cls, attributes):
        """Creation Axiom 3: Initial Status must be Pending"""
        status = attributes["status"]
        if status != GNodeStatus.Pending:
            raise DcError(
                f"Creation Axiom 3: Initial Status must be Pending. Got {status}"
            )

    @classmethod
    def _creation_axiom_4(cls, attributes):
        """Creation Axiom 4: On creation, PrevAlias is None"""
        prev_alias = attributes["prev_alias"]
        if prev_alias is not None:
            raise DcError(
                f"Creation Axiom 4: On creation, PrevAlias is None. Got {prev_alias}"
            )

    @classmethod
    def _creation_axiom_5(cls, attributes):
        """Creation Axiom 5: TOPOLOGY. Assume the Alias has at least two words.
        Then
          - EXISTENCE the parent GNode (from alias) must in the GNodeFactory, unless GNode is root.
          - STATUS
             - If status is NOT active, all children must be PermanentlyDeactivated.
          - ROLE
            - If role is Ctn or MarketMaker, the parent must be a root, or have either role Ctn or MarketMaker.
            - If the role is AtomicMeausurementNode or AtomicTNode, the parent must be either Ctn or MarketMaker.
            - If the role is TerminalAsset, the parent must be either AtomicMeasurementNode or AtomicTNode
        """
        alias: str = attributes["alias"]
        if len(alias.split(".")) == 1:
            "remaining axioms all have to do with parent-child relationship"
            return

        role = attributes["role"]
        status = attributes["status"]

        parent = cls.parent_from_alias(alias)
        if parent is None:
            raise DcError(f"Create Axiom 5: non-root {alias} must have Active parent!")
        if parent.status is GNodeStatus.PermanentlyDeactivated and not (
            status is GNodeStatus.PermanentlyDeactivated
        ):
            raise DcError(
                "Creation Axiom 5`: the parent of an PermanentlyDeactivated GNode must be PermanentlyDeactivated."
            )

        if (role is CoreGNodeRole.ConductorTopologyNode) or (
            role is CoreGNodeRole.MarketMaker
        ):
            if not (
                parent.is_root()
                or (parent.role is CoreGNodeRole.ConductorTopologyNode)
                or (parent.role is CoreGNodeRole.MarketMaker)
            ):
                raise DcError(
                    f"Creation Axiom 5`: the parent of a {role} must be ConductorTopologyNode"
                    f" or MarketMaker, not {parent.role}"
                )
        if (role is CoreGNodeRole.AtomicMeteringNode) or (
            role is CoreGNodeRole.AtomicTNode
        ):
            if not (
                (parent.role is CoreGNodeRole.ConductorTopologyNode)
                or (parent.role is CoreGNodeRole.MarketMaker)
            ):
                raise DcError(
                    f"Creation Axiom 5 `: the parent of a {role} must be ConductorTopologyNode"
                    f" or MarketMaker, not {parent.role}"
                )
        if role is CoreGNodeRole.TerminalAsset:
            if not (
                (parent.role is CoreGNodeRole.AtomicMeteringNode)
                or (parent.role is CoreGNodeRole.AtomicTNode)
            ):
                raise DcError(
                    f"Joint Axiom `: the parent of a {role} must be AtomicMete PringNode"
                    f" or AtomicTNode, not {parent.role}"
                )

    #########################################################
    # Schema axioms (shared with type)
    #########################################################

    @classmethod
    def _schema_axiom_1(cls, attributes: Dict):
        """Schema Axiom 1: The following attributes must exist: g_node_id,
        alias, status, role, g_node_registry_addr"""
        if "g_node_id" not in attributes.keys():
            raise DcError("g_node_id must exist")
        if "alias" not in attributes.keys():
            raise DcError("alias must exist")
        if "status" not in attributes.keys():
            raise DcError(f"status must exist for g node {attributes}")
        if "role" not in attributes.keys():
            raise DcError(f"role must exist for g node {attributes}")
        if "g_node_registry_addr" not in attributes.keys():
            raise DcError(f"g_node_registry_addr must exist for g node {attributes}")

    @classmethod
    def _schema_axiom_2(cls, attributes):
        """Schema Axiom 2: FORMATTING
        If they exist:
        - Alias must be a string of format LRD Alias
        - PrevAlias must be a string of format LRD Alias
        - GNodeId must be a string of format UuidCanoicalTextual
        - OwnershipDeedId must be an integer,
        - OwnershipDeedValidatorAddr must be a string of format AlgoAddressStringFormat
        - OwnerAddr must be a string of format AlgoAddressStringFormat
        - DaemonAddr must be a string of format AlgoAddressStringFormat
        - TradingRightsId must be an integer"""

        g_node_registry_addr = attributes["g_node_registry_addr"]
        ownership_deed_nft_id = attributes["ownership_deed_nft_id"]
        ownership_deed_validator_addr = attributes["ownership_deed_validator_addr"]
        owner_addr = attributes["owner_addr"]
        daemon_addr = attributes["daemon_addr"]
        trading_rights_id = attributes["trading_rights_id"]
        if ownership_deed_nft_id:
            if not isinstance(ownership_deed_nft_id, int):
                raise DcError("Schema Axiom 2: OwnershipDeedId must be an integer")
        if trading_rights_id:
            if not isinstance(trading_rights_id, int):
                raise DcError("Schema Axiom 2: TradingRightsId must be an integer ")
        for addr in [
            ownership_deed_validator_addr,
            owner_addr,
            daemon_addr,
            g_node_registry_addr,
        ]:
            if addr:
                if not isinstance(addr, str):
                    raise DcError(f"jSchema Axiom 2: {addr} must be a string")
                try:
                    check_is_algo_address_string_format(addr)
                except GwTypeError as e:
                    raise DcError(
                        f"Schema Axiom 2: {addr} must have format AlgoAddressStringFormat"
                    )

    @classmethod
    def _schema_axiom_3(cls, attributes):
        """Schema Axiom 3: Assume OwnershipDeedNftId exists and role is TerminalAsset.
        The Nft must be for a Valid TaDeed created by the 2-sig [GnfAdmin, ownership_deed_validator_addr]
        multi whose asset_name is this GNodeAlias."""
        pass

    @classmethod
    def _schema_axiom_4(cls, attributes):
        """Schema Axiom 4: Assume role is TerminalAsset and status is Active. Then the OwnershipDeedNftId
        must exist, and must be owned by the 2-sig [GnfAdmin, smart_daemon_addr, owner_addr] multi
        """
        pass

    #####################################
    # Joing axioms for creation or update
    #####################################

    @classmethod
    def _joint_axiom_1(cls, attributes):
        """Joint Axiom 1: There is at most one root GNode."""
        aliases = list(set(cls.by_alias.keys()).union({attributes["alias"]}))
        roots = list(filter(lambda x: len(x.split(".")) == 1, aliases))
        if len(roots) > 1:
            raise DcError(
                f"Joint Axiom 1: There is at most one root GNode. Already have {set(roots) -  {attributes['alias']} }"
            )

    @classmethod
    def _joint_axiom_2(cls, attributes):
        """Joint Axiom 2: If GpsPointId is not None, then  there is a GpsPoint in the
        GNodeFactory with that GpsPointId"""
        gps_point_id = attributes["gps_point_id"]
        pass
        if gps_point_id is not None:
            if gps_point_id not in GpsPoint.by_id.keys():
                raise DcError(
                    "Joint Axoim 2: If GpsPointId is not None, then it exists in the GpsPoints"
                    f"Got {gps_point_id} for {attributes['g_node_id']}"
                )

    ##########################
    # Object update
    ##########################

    def check_update_axioms(self, new_attributes):
        self._update_axiom_1(new_attributes)
        self._update_axiom_2(new_attributes)
        self._update_axiom_3(new_attributes)
        self._update_axiom_4(new_attributes)
        self._update_axiom_5(new_attributes)
        self._schema_axiom_1(new_attributes)
        self._schema_axiom_2(new_attributes)
        self._schema_axiom_3(new_attributes)
        self._schema_axiom_4(new_attributes)
        self._joint_axiom_1(new_attributes)
        self._joint_axiom_2(new_attributes)

    def _update_axiom_1(self, new_attributes):
        """Update Axiom 1: g_node_alias cannot be used, now or previously, by this or
        other BaseGNodes.
        """
        # Done in models.py, as it requires access to BaseGNodeHistory
        pass

    def _update_axiom_2(self, new_attributes):
        """Update Axiom 2: g_node_id is Immutable"""
        if self.g_node_id:
            if new_attributes["g_node_id"] != self.g_node_id:
                raise DcError(f"Update Axiom 1: g_node_id is Immutable.")

    def _update_axiom_3(self, new_attributes):
        """Update Axiom 3:  Status update rules:
        - Pending can only change to Active,
        - Active can only change to Suspended or PermanentlyDeactivated
        - Suspended can only change to Active or PermanentlyDeactivated
        - PermanentlyDeactivated cannot change."""
        new_status = new_attributes["status"]
        if self.status is GNodeStatus.Pending:
            if not (
                (new_status is GNodeStatus.Pending)
                or (new_status is GNodeStatus.Active)
            ):
                raise DcError(
                    f"Update Axiom 2: Pending can only change to Active, not {new_status}"
                )
        elif self.status is GNodeStatus.Active:
            if not (
                (new_status is GNodeStatus.Active)
                or (new_status is GNodeStatus.PermanentlyDeactivated)
                or (new_status is GNodeStatus.Suspended)
            ):
                raise DcError(
                    f"Update Axiom 2: Active can only change to Suspended or PermanentlyDeactivated, not {new_status}"
                )
        elif self.status is GNodeStatus.Suspended:
            if not (
                (new_status is GNodeStatus.Active)
                or (new_status is GNodeStatus.PermanentlyDeactivated)
                or (new_status is GNodeStatus.Suspended)
            ):
                raise DcError(
                    f"Update Axiom 2: Suspended can only change to Active or PermanentlyDeactivated, not {new_status}"
                )
        elif self.status is GNodeStatus.PermanentlyDeactivated:
            if not (new_status is GNodeStatus.PermanentlyDeactivated):
                raise DcError(
                    f"Update Axiom 3: Permanently Deactivated status cannot change. Got {new_status}"
                )

    def _update_axiom_4(self, attributes):
        """Update Axiom 4: TOPOLOGY. Assume the Alias has at least two words.
        Then
          - EXISTENCE the parent GNode  must in the GNodeFactory, unless GNode is root.
          - STATUS
             - If status is NOT Active, all children must have status PermanentlyDeactivated
            - If role is Ctn or MarketMaker, the parent must be a root, or have either role Ctn or MarketMaker.
            - If the role is AtomicMeausurementNode or AtomicTNode, the parent must be either Ctn or MarketMaker.
            - If the role is TerminalAsset, the parent must be either AtomicMeasurementNode or AtomicTNode
        """
        alias: str = attributes["alias"]
        role = attributes["role"]
        status = attributes["status"]
        if len(alias.split(".")) == 1:
            is_root = True
        else:
            is_root = False

        # If status is NOT Active, all children must have status PermanentlyDeactivated
        # LOOKING UP AT PARENT
        if not is_root:
            parent = self.parent()
            if parent is None and self.status != GNodeStatus.PermanentlyDeactivated:
                raise DcError(
                    f"Update Axiom 4: non-root {alias} with status {self.status} must have Active parent!"
                )

        # LOOKING DOWN AT DESCENDANTS
        if status != GNodeStatus.Active:
            children_status_set = set(filter(lambda x: x.status, self.children()))
            if (
                children_status_set != {GNodeStatus.PermanentlyDeactivated}
                and children_status_set != set()
            ):
                raise DcError(
                    f"Cannot change role to {role} unless there are no children or "
                    f"children are all PermanentlyDeactivated. Children: {self.children()}"
                )

        if not is_root:
            if (role is CoreGNodeRole.ConductorTopologyNode) or (
                role is CoreGNodeRole.MarketMaker
            ):
                if not (
                    parent.is_root()
                    or (parent.role is CoreGNodeRole.ConductorTopologyNode)
                    or (parent.role is CoreGNodeRole.MarketMaker)
                ):
                    raise DcError(
                        f"Update Axiom 4`: the parent of a {role} must be ConductorTopologyNode"
                        f" or MarketMaker, not {parent.role}"
                    )
            if (role is CoreGNodeRole.AtomicMeteringNode) or (
                role is CoreGNodeRole.AtomicTNode
            ):
                if not (
                    (parent.role is CoreGNodeRole.ConductorTopologyNode)
                    or (parent.role is CoreGNodeRole.MarketMaker)
                ):
                    raise DcError(
                        f"Update Axiom 4`: the parent of a {role} must be ConductorTopologyNode"
                        f" or MarketMaker, not {parent.role}"
                    )
            if role is CoreGNodeRole.TerminalAsset:
                if not (
                    (parent.role is CoreGNodeRole.AtomicMeteringNode)
                    or (parent.role is CoreGNodeRole.AtomicTNode)
                ):
                    raise DcError(
                        f"Update Axiom 4`: the parent of a {role} must be AtomicMete PringNode"
                        f" or AtomicTNode, not {parent.role}"
                    )
        # LOOKING DOWN AT DESCENDANTS
        # Question: add the above again, or rely on role update axioms??
        # ... should probably add above again in case role update axioms change ...

    def _update_axiom_5(self, attributes):
        """Update Axiom 5: Role update rules.  TerminalAsset, InterconnectionComponent, and
        Other cannot change roles. ConductorTopologyNode can turn into MarketMaker and vice
        versa. AtomicMeteringNode can turn into AtomicTNode and vice versa"""
        pass

    def _update_axiom_6(self, attributes):
        """Update Axiom 6: If alias has changed, then new prev_alias must equal
        original alias prior to the change. If alias has NOT changed then new prev_alias
         must equal original prev_alias."""
        new_alias: str = attributes["alias"]
        prev_alias: str = attributes["prev_alias"]
        if new_alias != self.alias:
            if prev_alias != self.alias:
                raise DcError(
                    f"Update Axiom 6: If alias has changed, then new prev_alias "
                    "must equal original alias prior to the change."
                )
        else:
            if prev_alias != self.prev_alias:
                raise DcError(
                    f"Update Axiom 6: If alias has NOT changed then new "
                    "prev_alias must equal original prev_alias."
                )


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_algo_address_string_format(v: str) -> None:
    """
    Checks if a given string is in the correct format for an Algorand address.

    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not in AlgoAddressStringFormat format

    Original (requires algosdk package):
    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")

    Sample:
    import algosdk
    private_key, address = algosdk.account.generate_account()
    # address will look something like '7XGIDZ7SJZJHPFGOOG7OQJ4OBCTV6TIT3HI4TFWKPPQ7PP2X3R22QKXN4U'
    """
    return



"""Type sim.timestep, version 000"""

import json
import logging
from typing import Any
from typing import Dict
from typing import Literal

from gw.errors import GwTypeError
from gw.utils import is_pascal_case
from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator
from pydantic.alias_generators import to_pascal
from pydantic.alias_generators import to_snake


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class SimTimestep(BaseModel):
    """
    Sent by TimeCoordinators to coordinate time.

    For simulated actors, time progresses discretely on receipt of these time steps.
    """

    from_g_node_alias: str = Field(
        title="The GNodeAlias of the sender",
        description="The sender should always be a GNode Actor of role TimeCoordinator.",
    )
    from_g_node_instance_id: str = Field(
        title="The GNodeInstanceId of the sender",
    )
    time_unix_s: int = Field(
        title="Current time in unix seconds",
    )
    timestep_created_ms: int = Field(
        title="The real time created, in unix milliseconds",
    )
    message_id: str = Field(
        title="MessageId",
    )
    type_name: Literal["sim.timestep"] = "sim.timestep"
    version: Literal["000"] = "000"

    class Config:
        populate_by_name = True
        alias_generator = to_pascal

    @field_validator("from_g_node_alias")
    def _check_from_g_node_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeAlias failed LeftRightDot format validation: {e}"
            )
        return v

    @field_validator("from_g_node_instance_id")
    def _check_from_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"FromGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @field_validator("time_unix_s")
    def _check_time_unix_s(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_s(v)
        except ValueError as e:
            raise ValueError(
                f"TimeUnixS failed ReasonableUnixTimeS format validation: {e}"
            )
        return v

    @field_validator("timestep_created_ms")
    def _check_timestep_created_ms(cls, v: int) -> int:
        try:
            check_is_reasonable_unix_time_ms(v)
        except ValueError as e:
            raise ValueError(
                f"TimestepCreatedMs failed ReasonableUnixTimeMs format validation: {e}"
            )
        return v

    @field_validator("message_id")
    def _check_message_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"MessageId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    def as_dict(self) -> Dict[str, Any]:
        """
        Translate the object into a dictionary representation that can be serialized into a
        sim.timestep.000 object.

        This method prepares the object for serialization by the as_type method, creating a
        dictionary with key-value pairs that follow the requirements for an instance of the
        sim.timestep.000 type. Unlike the standard python dict method,
        it makes the following substantive changes:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.
        """
        d = {
            to_pascal(key): value
            for key, value in self.model_dump().items()
            if value is not None
        }
        return d

    def as_type(self) -> bytes:
        """
        Serialize to the sim.timestep.000 representation.

        Instances in the class are python-native representations of sim.timestep.000
        objects, while the actual sim.timestep.000 object is the serialized UTF-8 byte
        string designed for sending in a message.

        This method calls the as_dict() method, which differs from the native python dict()
        in the following key ways:
        - Enum Values: Translates between the values used locally by the actor to the symbol
        sent in messages.
        - - Removes any key-value pairs where the value is None for a clearer message, especially
        in cases with many optional attributes.

        It also applies these changes recursively to sub-types.

        Its near-inverse is SimTimestep.type_to_tuple(). If the type (or any sub-types)
        includes an enum, then the type_to_tuple will map an unrecognized symbol to the
        default enum value. This is why these two methods are only 'near' inverses.
        """
        json_string = json.dumps(self.as_dict())
        return json_string.encode("utf-8")

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class SimTimestep_Maker:
    type_name = "sim.timestep"
    version = "000"

    @classmethod
    def tuple_to_type(cls, tuple: SimTimestep) -> bytes:
        """
        Given a Python class object, returns the serialized JSON type object.
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: bytes) -> SimTimestep:
        """
        Given a serialized JSON type object, returns the Python class object.
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise GwTypeError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise GwTypeError(f"Deserializing <{t}> must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> SimTimestep:
        """
        Deserialize a dictionary representation of a sim.timestep.000 message object
        into a SimTimestep python object for internal use.

        This is the near-inverse of the SimTimestep.as_dict() method:
          - Enums: translates between the symbols sent in messages between actors and
        the values used by the actors internally once they've deserialized the messages.
          - Types: recursively validates and deserializes sub-types.

        Note that if a required attribute with a default value is missing in a dict, this method will
        raise a GwTypeError. This differs from the pydantic BaseModel practice of auto-completing
        missing attributes with default values when they exist.

        Args:
            d (dict): the dictionary resulting from json.loads(t) for a serialized JSON type object t.

        Raises:
           GwTypeError: if the dict cannot be turned into a SimTimestep object.

        Returns:
            SimTimestep
        """
        for key in d.keys():
            if not is_pascal_case(key):
                raise GwTypeError(f"Key '{key}' is not PascalCase")
        d2 = dict(d)
        if "FromGNodeAlias" not in d2.keys():
            raise GwTypeError(f"dict missing FromGNodeAlias: <{d2}>")
        if "FromGNodeInstanceId" not in d2.keys():
            raise GwTypeError(f"dict missing FromGNodeInstanceId: <{d2}>")
        if "TimeUnixS" not in d2.keys():
            raise GwTypeError(f"dict missing TimeUnixS: <{d2}>")
        if "TimestepCreatedMs" not in d2.keys():
            raise GwTypeError(f"dict missing TimestepCreatedMs: <{d2}>")
        if "MessageId" not in d2.keys():
            raise GwTypeError(f"dict missing MessageId: <{d2}>")
        if "TypeName" not in d2.keys():
            raise GwTypeError(f"TypeName missing from dict <{d2}>")
        if "Version" not in d2.keys():
            raise GwTypeError(f"Version missing from dict <{d2}>")
        if d2["Version"] != "000":
            LOGGER.debug(
                f"Attempting to interpret sim.timestep version {d2['Version']} as version 000"
            )
            d2["Version"] = "000"
        d3 = {to_snake(key): value for key, value in d2.items()}
        return SimTimestep(**d3)


def check_is_left_right_dot(v: str) -> None:
    """Checks LeftRightDot Format

    LeftRightDot format: Lowercase alphanumeric words separated by periods, with
    the most significant word (on the left) starting with an alphabet character.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate <{v}> into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(
            f"Most significant word of <{v}> must start with alphabet char."
        )
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of <{v}> split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of <{v}> must be lowercase.")


def check_is_reasonable_unix_time_ms(v: int) -> None:
    """Checks ReasonableUnixTimeMs format

    ReasonableUnixTimeMs format: unix milliseconds between Jan 1 2000 and Jan 1 3000

    Args:
        v (int): the candidate

    Raises:
        ValueError: if v is not ReasonableUnixTimeMs format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp_ms = int(start_date.timestamp() * 1000)
    end_timestamp_ms = int(end_date.timestamp() * 1000)

    if v < start_timestamp_ms:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp_ms:
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_reasonable_unix_time_s(v: int) -> None:
    """Checks ReasonableUnixTimeS format

    ReasonableUnixTimeS format: unix seconds between Jan 1 2000 and Jan 1 3000

    Args:
        v (int): the candidate

    Raises:
        ValueError: if v is not ReasonableUnixTimeS format
    """
    from datetime import datetime
    from datetime import timezone

    start_date = datetime(2000, 1, 1, tzinfo=timezone.utc)
    end_date = datetime(3000, 1, 1, tzinfo=timezone.utc)

    start_timestamp = int(start_date.timestamp())
    end_timestamp = int(end_date.timestamp())

    if v < start_timestamp:
        raise ValueError(f"{v} must be after Jan 1 2000")
    if v > end_timestamp:
        raise ValueError(f"{v} must be before Jan 1 3000")


def check_is_uuid_canonical_textual(v: str) -> None:
    """Checks UuidCanonicalTextual format

    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Args:
        v (str): the candidate

    Raises:
        ValueError: if v is not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"<{v}> split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of <{v}> are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"<{v}> word lengths not 8-4-4-4-12")



"""Settings for the GNodeFactory, readable from environment and/or from env files."""

import datetime

from pydantic import SecretStr
from pydantic import field_validator
from pydantic import model_validator
from pydantic_settings import BaseSettings
from typing_extensions import Self

from gwbase.config.algo_settings import AlgoApiSecrets
from gwbase.config.algo_settings import Public
from gwbase.config.rabbit_settings import RabbitBrokerClient
from gwbase.config.utils import check_g_node_alias
from gwbase.config.utils import check_is_algo_secret_key_format
from gwbase.config.utils import check_is_reasonable_unix_time_s
from gwbase.enums import GNodeRole
from gwbase.enums import UniverseType


DEFAULT_ENV_FILE = ".env"


class GNodeSettings(BaseSettings):
    """
    Template settings for a GNode.
    """

    rabbit: RabbitBrokerClient = RabbitBrokerClient()
    g_node_alias: str = "d1.isone.unknown.gnode"
    g_node_id: str = "e23eb2ec-4064-4921-89d4-b006edc81216"
    g_node_instance_id: str = "00000000-0000-0000-0000-000000000000"
    my_super_alias: str = "d1.super1"
    g_node_role_value: str = "GNode"
    sk: SecretStr = SecretStr(
        "3g+IYDCVM84Ady7a8fGImRkEZ77+a4e3i14ub0QMjM/JKlzB2GNdv0S+lqMsYgPiGbd7aAp5943X5NzvdQJohw=="
    )
    # Secret key for public Algorand adress ZEVFZQOYMNO36RF6S2RSYYQD4IM3O63IBJ47PDOX4TOO65ICNCDUF4HMLE
    universe_type_value: str = "Dev"
    my_time_coordinator_alias: str = "d1.time"
    initial_time_unix_s: int = int(
        datetime.datetime(2020, 1, 1, 4, 20, tzinfo=datetime.timezone.utc).timestamp()
    )

    log_level: str = "INFO"
    public: Public = Public()
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    redis_endpoint: str = "localhost"
    minute_cron_file: str = "cron_last_minute.txt"
    hour_cron_file: str = "cron_last_hour.txt"
    day_cron_file: str = "cron_last_day.txt"

    @field_validator("initial_time_unix_s", mode="before")
    def convert_initial_time_unix_s(cls, value):
        return int(value)

    @model_validator(mode="after")
    def validate_setting_axioms(self) -> Self:
        """
        Validates the following:
            - universe_type_value belongs to the GridWorks UniverseType enum
            - g_node_role_value belongs to the GridWOrks GNodeRole enum
            - All GNodeAliases (self, my time coordinator, my super) have the correct
            LeftRightDot format and match the alias pattern for the universe type
            - sk.get_secret_value() has the format of an Algorand secret key
            - initial_time_unix_s is a reasonable unix time in ms
        """
        if self.universe_type_value not in UniverseType.values():
            raise ValueError("universe_type_value must belong to UniverseType.values()")
        universe_type = UniverseType(self.universe_type_value)
        if self.g_node_role_value not in GNodeRole.values():
            raise ValueError("g_node_role_value must belong to GNodeRole.values()")
        check_g_node_alias(alias=self.g_node_alias, universe_type=universe_type)
        check_g_node_alias(
            alias=self.my_time_coordinator_alias, universe_type=universe_type
        )
        check_g_node_alias(alias=self.my_super_alias, universe_type=universe_type)
        check_is_algo_secret_key_format(self.sk.get_secret_value())
        check_is_reasonable_unix_time_s(self.initial_time_unix_s)
        return self

    class Config:
        env_prefix = "GNODE_"
        env_nested_delimiter = "__"


class SupervisorSettings(BaseSettings):
    g_node_alias: str = "d1.isone.ver.keene.super1"
    g_node_id: str = "664a3250-ce51-4fe3-9ce9-a4b6416451fb"
    g_node_instance_id: str = "20e7edec-05e5-4152-bfec-ec21ddd2e3dd"
    supervisor_container_id: str = "995b0334-9940-424f-8fb1-4745e52ba295"
    g_node_role_value: str = "Supervisor"
    my_time_coordinator_alias: str = "d1.time"
    log_level: str = "INFO"
    universe_type_value: str = "Dev"
    world_instance_name: str = "d1__1"
    rabbit: RabbitBrokerClient = RabbitBrokerClient()

    class Config:
        env_prefix = "SUPER_"
        env_nested_delimiter = "__"



import algosdk
from pydantic import BaseModel
from pydantic import SecretStr
from pydantic import field_validator
from pydantic import model_validator
from pydantic_settings import BaseSettings


class Public(BaseModel):
    """
    Publicly available information about the GNodeFactory, including:
      - GnfAdminAddr
      - GnfApiRoot
      - TaValidatorFundingThresholdAlgos
      - TaDeedConsiderationAlgos

    Also includes useful information shartcuts for running the simulated Millinocket
    demo. In this demo there is only one MarketMaker, and one TaValidator.
    Public Algorand addresses and ApiRoots are included for both.
    """

    gnf_api_root: str = "http://localhost:8000"
    dev_market_maker_api_root: str = "http://localhost:7997"
    dev_ta_validator_api_root: str = "http://localhost:8001"
    gnf_admin_addr: str = "RNMHG32VTIHTC7W3LZOEPTDGREL5IQGK46HKD3KBLZHYQUCAKLMT4G5ALI"
    gnr_addr: str = "X2ASUAUPK5ICMGDXQZQKBPSXWEJLBA4KKQ2TXW2KWO2JQTLY3J2Q4S33WE"
    dev_market_maker_addr: str = (
        "JMEUH2AXM6UGRJO2DBZXDOA2OMIWQFNQZ54LCVC4GQX6QDOX5Z6JRGMWFA"
    )
    dev_ta_validator_addr: str = (
        "7QQT4GN3ZPAQEFCNWF5BMF7NULVK3CWICZVT4GM3BQRISD52YEDLWJ4MII"
    )
    dev_ta_validator_multi_addr: str = (
        "Y5TRQXIJHWJ4OHCZSWP4PZTCES5VWOF2KDTNYSMU5HLAUXBFQQDX6IR5KM'"
    )
    ta_validator_funding_threshold_algos: int = 100
    ta_deed_consideration_algos: int = 50
    universe: str = "dev"
    gnf_graveyard_addr: str = (
        "COA6SYUOBE33F5JDYEGC5XAD43QRG3VGHNNQXLYWFSSQEHDQ5HJ52NDNPI"
    )
    algod_address: str = "http://localhost:4001"
    kmd_address: str = "http://localhost:4002"
    gen_kmd_wallet_name: str = "unencrypted-default-wallet"


class AlgoApiSecrets(BaseModel):
    algod_token: SecretStr = SecretStr(
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )
    kmd_token: SecretStr = SecretStr(
        "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )
    gen_kmd_wallet_password: SecretStr = SecretStr("")


class VanillaSettings(BaseSettings):
    algo_api_secrets: AlgoApiSecrets = AlgoApiSecrets()
    public: Public = Public()

    class Config:
        env_prefix = "VANILLA_"
        env_nested_delimiter = "__"



"""
GridWorks Enums used in gwbase, the Application Shared Language (ASL) used by SCADA
devices and AtomicTNodes to communicate with each other. These enums play a specific structural
role as semantic "glue" within ASLs.

Key attributes:
  - Enum values are translated into "GridWorks Type Enum Symbols" (GtEnumSymbols) when embedded
  in a serialized type sent as a message from one Application and/or Actor to another.
  - Each Enum has a unique name in the type registry (like spaceheat.telemetry.name), along
  with a version (like 001).
  - That name are interpretted locally in the SDK and do not necessarily carry the larger
  context of the unique type registry name (for example gwbase uses TelemetryName, since
  the `spaceheat` context goes without saying).
  - Each Value/Symbol pair also has a version. Value/Symbol pairs cannot be changed or removed.
  The only adjustments that can be made to an enum are adding more Value/Symbols. This is to
  support forwards- and backwards- compatability in GridWorks Types that use these enums.

If Enums are "glue", then GridWorks Types are the building blocks of SALs. Every SAL is comprised
of a set of shared GridWorks Types.

Application Shared Languages are an evolution of the concept of Application Programming Interfaces.
In a nutshell, an API can be viewed as a rather restricted version of an SAL, where only one application
has anything complex/interesting to say and, in general, the developers/owners of that application
have sole responsibility for managing the versioning and changing of that API. Note also that SALs
do not make any a priori assumption about the relationship (i.e. the default client/server for an API)
or the message delivery mechanism (i.e. via default GET/POST to RESTful URLs). For more information
on these ideas:
  - [GridWorks Enums](https://gridwork-type-registry.readthedocs.io/en/latest/types.html)
  - [GridWorks Types](https://gridwork-type-registry.readthedocs.io/en/latest/types.html)
  - [ASLs](https://gridwork-type-registry.readthedocs.io/en/latest/asls.html)
 """

from gwbase.enums.core_g_node_role import CoreGNodeRole
from gwbase.enums.g_node_role import GNodeRole
from gwbase.enums.g_node_status import GNodeStatus
from gwbase.enums.gni_status import GniStatus
from gwbase.enums.message_category import MessageCategory
from gwbase.enums.message_category_symbol import MessageCategorySymbol
from gwbase.enums.strategy_name import StrategyName
from gwbase.enums.supervisor_container_status import SupervisorContainerStatus
from gwbase.enums.universe_type import UniverseType


__all__ = [
    "CoreGNodeRole",  # [core.g.node.role.001](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#coregnoderole)
    "GNodeRole",  # [g.node.role.001](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#gnoderole)
    "GNodeStatus",  # [g.node.status.100](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#gnodestatus)
    "GniStatus",  # [gni.status.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#gnistatus)
    "MessageCategory",  # [message.category.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#messagecategory)
    "MessageCategorySymbol",  # [message.category.symbol.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#messagecategorysymbol)
    "StrategyName",  # [strategy.name.001](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#strategyname)
    "SupervisorContainerStatus",  # [supervisor.container.status.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#supervisorcontainerstatus)
    "UniverseType",  # [universe.type.000](https://gridworks-type-registry.readthedocs.io/en/latest/enums.html#universetype)
]



from typing import Dict
from typing import Optional

from gwbase.enums import GNodeRole
from gwbase.enums import StrategyName


class GNodeStrategy:
    by_id: Dict[StrategyName, "GNodeStrategy"] = {}

    def __new__(cls, name: StrategyName, *args, **kwargs) -> "GNodeStrategy":  # type: ignore
        try:
            return cls.by_id[name]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[name] = instance
            return instance

    def __init__(
        self,
        name: StrategyName,
        role: GNodeRole,
        contact_email: str,
        description: Optional[str] = "",
        is_private: bool = False,
        repo_url: Optional[str] = "None",
    ):
        self.name = name
        self.role = role
        self.contact_email = contact_email
        self.description = description
        self.is_private = is_private
        self.repo_url = repo_url

    def __repr__(self) -> str:
        s = (
            f"{self.role} GNodeStrategy {self.name.value}: Is Private: {self.is_private}, Repo: {self.repo_url}, contact: "
            f"{self.contact_email}\n {self.description}"
        )
        return s


NoActor = GNodeStrategy(
    name=StrategyName.NoActor,
    role=GNodeRole.GNode,
    contact_email="gridworks.gridworks-consulting.com",
    is_private=False,
    description="The NoActor strategy is a placeholder for a GNode that is passive and has no actor.",
)

SupervisorA = GNodeStrategy(
    name=StrategyName.SupervisorA,
    role=GNodeRole.Supervisor,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-atn",
    is_private=False,
    description="A simple supervisor that monitors its supervised AtomicTNode GNode strategies"
    "via a heartbeat health check. [More Info](https://gridworks.readthedocs.io/en/latest/supervisor.html)",
)

WorldA = GNodeStrategy(
    name=StrategyName.WorldA,
    role=GNodeRole.World,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-world",
    is_private=False,
    description="Authority on GNodeInstances, and their private keys. Maintains a FastAPI used "
    "for relational information about backoffice information held locally"
    "and/or in the GNodeRegistry/GNodeFactory. [More Info](https://gridworks.readthedocs.io/en/latest/world-role.html)",
)

AtnHeatPumpWithBoostStore = GNodeStrategy(
    name=StrategyName.AtnHeatPumpWithBoostStore,
    role=GNodeRole.AtomicTNode,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-atn-spaceheat",
    is_private=True,
    description="AtomicTNode for a heat pump thermal storage heating system with a boost element and a thermal "
    "heated by the boost element. "
    "[More on AtomicTNodes](https://gridworks.readthedocs.io/en/latest/atomic-t-node.html)",
)

TcGlobalA = GNodeStrategy(
    name=StrategyName.TcGlobalA,
    role=GNodeRole.TimeCoordinator,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-timecoordinator",
    is_private=False,
    description="Used to manage the global time of the Gridworks system when run with"
    "in a fully simulated universe. "
    "[More on TimeCoordinators](https://gridworks.readthedocs.io/en/latest/time-coordinator.html)",
)

MarketMakerA = GNodeStrategy(
    name=StrategyName.MarketMakerA,
    role=GNodeRole.MarketMaker,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-marketmaker",
    is_private=False,
    description="Runs a two-sided market associated to its GNode as part of the copper"
    "GNode sub-tree. "
    "[More on MarketMakers](https://gridworks.readthedocs.io/en/latest/market-maker.html)",
)

AtnBrickStorageHeater = GNodeStrategy(
    name=StrategyName.AtnBrickStorageHeater,
    role=GNodeRole.AtomicTNode,
    contact_email="gridworks.gridworks-consulting.com",
    repo_url="https://github.com/thegridelectric/gridworks-atn",
    is_private=False,
    description="Publicly available Dijkstra-based AtomicTNode strategy for a brick storage"
    "heater. These heaters are rooom units that store heat in a brick core, are "
    "heated with resistive elements, and typically have a fan to blow air over the"
    "brick core. They are sometimes called Electric Thermal Storage (ETS) heaters,"
    "and in the UK are often called Economy 7 heaters or Night Storage Heaters."
    "A strategy very similar to this was used by VCharge to manage an ETS fleet"
    "of several thousand heaters in Pennsylvania."
    "This strategy is meant to serve as a template for other private strategies,"
    "and also allows for an end-to-end simulation of a realistic aggregated"
    "transactive load capable of generating a highly elastic bid curve"
    "[More Info](https://gridworks-atn.readthedocs.io/en/latest/brick-storage-heater.html)",
)



import logging
from typing import Dict

from gwbase.enums import SupervisorContainerStatus


LOG_FORMAT = (
    "%(levelname) -10s %(asctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


class SupervisorContainer:
    by_id: Dict[str, "SupervisorContainer"] = {}

    def __new__(cls, supervisor_container_id: str, *args, **kwargs) -> "SupervisorContainer":  # type: ignore
        try:
            return cls.by_id[supervisor_container_id]
        except KeyError:
            instance = super().__new__(cls)
            cls.by_id[supervisor_container_id] = instance
            return instance

    def __init__(
        self,
        supervisor_container_id: str,
        status: SupervisorContainerStatus,
        world_instance_name: str,
        supervisor_g_node_instance_id: str,
        supervisor_g_node_alias: str,
    ):
        self.supervisor_container_id = supervisor_container_id
        self.status = status
        self.world_instance_name = world_instance_name
        self.supervisor_g_node_instance_id = supervisor_g_node_instance_id
        self.supervisor_g_node_alias = supervisor_g_node_alias
